%%%% Data centre system model, as described in Tatchell-Evans (2017,
%%%% Chapter 5)

VIn=6.9912; % Set equal to supply flow rate in corresponding CFD model
EServ=82.06498; % Set equal to total server power consumption in corresponding CFD model
EServ_0=79.353; % Server power consumption required at zero external static pressure
VServ_0=79.353/(1.005*12.5); % Server flow rate required at zero external static pressure

ServerInletTemps2=293.1; % Set equal to max server inlet temperature in corresponding CFD model
T_0=293; % Default supply temperature

T_PA_Sup=27-(max(max(ServerInletTemps2(:,:)))-T_0); % Determines the supply temperature required to ensure the maximum inlet temperature does not exceed ASHRAE recommendations
T_amb=11; % Mean temperature for London=11C

dT_IT=12.5; % Default temperature rise across servers
c_pAir=1.005; % specific heat capacity of air (kJ/kgK)
c_pWater=4.18; % specific heat capacity of water (kJ/kgK)
rho_Water=1000; % density of water (kg/m^3)
rho_Air=1.2; % density of air (kg/m^3)
eta_CRACFan=0.9; % Efficiency of CRAC fans, consistent with Salim & Tozer
eta_CWPump=0.9; % Efficiency of chilled water pumps, consistent with Salim & Tozer
CWPumpPowerRef=2*60.91*0.0014/eta_CWPump; % kW. Uniflair TD/UCV 0700 has a pressure drop of 60.91 kPa at 1.4 l/s.  Assume same pressure drop in the economiser, hence factor of 2.
h_CRACRef=23.9/7.27; % The Uniflair TD/UCV achieves a sensible heat transfer of 23.9kW at LMTD=7.27
Q_CWRef=0.0014*sum(EServ_0)*1.1/31.2; % referenced Uniflair data corresponds to CW flow rate of 1.4 l/s, providing 31.2kW of cooling.  Assume heat exchangers etc. are scaled up appropriately, to cool the housed IT at a PUE of 1.1.
CWPumpPowerRef=CWPumpPowerRef*sum(EServ_0)*1.1/31.2; % Assume reference CW pump power scales linearly from Uniflair data for condition where load on chilled water loop is 1.1x IT load.  Will then scale from there by the cube of flow rate.
index=0.631; % index for relationship between h_CRAC and Reynolds number (figures from the literature range from 0.581-0.681)

T_PA_Ret=T_PA_Sup+EServ/(VIn*rho_Air*c_pAir); % Temperature (K) of process air returning to CRAC
Q_CRACRef=1.2*sum(sum(VServ_0)); % For calculating CRAC fan power consumption, assuming a default system with 20% bypass.
CRACFanPower=(Q_CRACRef/1.66)*2.2*((VIn/Q_CRACRef)^3); % Power consumption of CRAC fans (kW).  Uniflair TD/UCV 0700 fan uses 2.2kW at 1.66m^3/s.  Scale this with the reference CRAC flow rate to get reference CRAC fan power, then scale with cube of flow rate for variation.
CRACLoad=EServ;
h_CRACRef=h_CRACRef*Q_CRACRef/1.66; % Assume heat transfer per LMTD scales linearly with size of CRAC

% heat exchanger calculations
h_CRAC=h_CRACRef*(VIn/Q_CRACRef)^index;
% Now test over a range of dT_CW
i=0;
V_CWControl=0; % This variable simply records whether or not CW flow rate was controlled to minimise power consumption
Q_EcoRef=Q_CRACRef*2;

for dT_CW=1:10 % for CW temperature rise of 1 to 10
    i=i+1;
    Q_CW(i)=(CRACLoad+CRACFanPower*(1-eta_CRACFan))/(dT_CW*c_pWater*rho_Water); % calculate minimum CW flow rate required to achieve the heat transfer at this CW dT
    
    CWPumpPower(i)=CWPumpPowerRef*(Q_CW(i)/Q_CWRef)^3; % calculate CW pump power assuming it scales with cube of flow rate
    
    LMTD_required(i)=(CRACLoad+CRACFanPower*(1-eta_CRACFan))/h_CRAC; % This LMTD is required to achieve the heat transfer on the air side
    % Finding CW supply temperature requires numerical methods:
    %T_CWh=T_CWc+dT_CW;
    %dTa=T_PA_Ret-T_CWh;    
    digitsOld=digits;
    digits(6)
    syms T_CWc
    T_CWc=double(vpasolve((T_PA_Ret-(T_CWc+dT_CW)-(T_PA_Sup-T_CWc))/log((T_PA_Ret-(T_CWc+dT_CW))/(T_PA_Sup-T_CWc)) == LMTD_required(i), T_CWc));
    digits(digitsOld)
    T_CWcStore(i)=T_CWc;
    
    EcoLoad(i)=CRACLoad+CRACFanPower*(1-eta_CRACFan)+CWPumpPower(i)*(1-eta_CWPump); % Load on the economiser
    dTb=T_CWc-T_amb;
    % Noting that h_eco=h_CRACRef * ((Q_CW/Q_CWRef)^0.8) * (((Q_Eco)/Q_PARef)^index)
    % and T_EA_Ex=T_amb+EcoLoad/(rho_Air*Q_Eco*c_pAir)
    % also EcoLoad=h_eco(LMTD_eco)
    % LMTD_eco=(((T_CWc+dT_CW)-(T_amb+EcoLoad/(rho_Air*Q_EA*c_pAir)))-dTb)/log(((T_CWc+dT_CW)-(T_amb+EcoLoad/(rho_Air*Q_Eco*c_pAir)))/dTb)
    digitsOld=digits;
    digits(6)
    syms Q_Eco
    % Combining the above equations allows us to solve numerically for the
    % flow rate required of the economiser fans:
    Q_Eco=double(vpasolve((h_CRACRef*((Q_CW(i)/Q_CWRef)^0.8)*(((Q_Eco)/Q_EcoRef)^index)) * (((T_CWc+dT_CW)-(T_amb+EcoLoad(i)/(rho_Air*Q_Eco*c_pAir)))-dTb)/log(((T_CWc+dT_CW)-(T_amb+EcoLoad(i)/(rho_Air*Q_Eco*c_pAir)))/dTb) == EcoLoad(i), T_CWc));
    for j=0.4:0.2:1.6
        Q_Eco1(round((j-0.2)/0.2))=Q_EcoRef*j;
    end
    for j=1:7
        HeatRejection(i,j)=(h_CRACRef*((Q_CW(i)/Q_CWRef)^0.8)*(((Q_Eco1(j))/Q_EcoRef)^index)) * (((T_CWc+dT_CW)-(T_amb+EcoLoad(i)/(rho_Air*Q_Eco1(j)*c_pAir)))-dTb)/log(((T_CWc+dT_CW)-(T_amb+EcoLoad(i)/(rho_Air*Q_Eco1(j)*c_pAir)))/dTb);
    end
    digits(digitsOld)
    if isempty(Q_Eco)==1
        Q_EcoStore(i)=9999999;
        EcoFanPower(i)=9999999;
        Power(i)=9999999;
    else
        Q_EcoStore(i)=Q_Eco;
        EcoFanPower(i)=(Q_EcoRef/1.66)*2.2*((Q_Eco/Q_EcoRef)^3); % Power consumption of economiser fans (kW).  Uniflair TD/UCV 0700 fan uses 2.2kW at 1.66m^3/s.  Scale this with the reference CRAC flow rate to get reference fan power, then scale with cube of flow rate for variation.
        Power(i)=[EcoFanPower(i)+CWPumpPower(i)];
    end
    V_CWControl=1; % Setting this variable to 1 indicates that CW flow rate was controlled to minimise power consumption
end

[MinPower i]=min(Power); % find index for combined power consumption of CW pump and economiser fan

TotalPower=MinPower+CRACFanPower+EServ;
COP_C=CRACLoad/(TotalPower-EServ-CRACFanPower)
PUE=TotalPower/EServ
Results=vertcat([EServ CRACFanPower EcoFanPower(i) CWPumpPower(i) TotalPower PUE]);
csvwrite('Results.csv',Results);