 % PF-CFD model of Test Data Centre, as described in Tatchell-Evans 2017, chapter 5.

clear all
Counter=0;

TimeStart=[datestr(now, 'HH') datestr(now, 'MM')]

PhiIt=2;

% Fix supply flow rate or p_CA?
prompt='Fix supply flow rate or cold aisle pressurisation? 1 for flow rate, 2 for pressure.';
EquilCond=input(prompt);
if EquilCond==1
    prompt='Set supply flow rate in m^3/s';
    VIn=input(prompt);
else
    prompt='Set pressure drop from cold to hot aisle in Pa';
    p_CA=input(prompt);
end


%% Properties of air etc., bypass and server flows, supply air conditions
rho_0=1.2; % kg/m^3
c_p=1.005; % kJ/kg.K
k_h=0.0264*10^-3; % Heat transfer coefficient, kW/mK at 17C and 100kPa according to Perry's
p_0=101000; % Atmospheric pressure, Pa
T_0=293; % Supply temperature, K

%% Dimensions etc.
% Original cell numbers
N_racks=4;
CellSide=0.8/8;
nx=round(N_racks*0.8/CellSide);
ny=round(3.4/CellSide);
nz=round(3.97/CellSide);
L=nx*CellSide;
H=ny*CellSide;
D=nz*CellSide;

Dx=L/nx; % mesh increment in the x direction (m)
Dy=H/ny; % mesh increment in the y direction (m)
Dz=D/nz; % mesh increment in the z direction (m)

%%%%% For test data centre inlet
AInlet=pi*(0.315/2)^2; % Area of inlet
HInlet=AInlet^0.5; % Height of inlet
DInlet=AInlet^0.5; % Width of inlet
Inlety=0.845; %y-offset of inlet
Inletz=0.665; % z-offset of inlet
% InletAngle=32*2*pi/360; % Angle of inlet flow (Radians)

%%%%%%% Circular outlet in HA
HOutlet=AInlet^0.5; % Height of outlet
DOutlet=AInlet^0.5; % Width of outlet
WOutlet=0; % Depth of outlet
Outletx=0; % x-offset of outlet
Outlety=0.845; %y-offset of outlet
Outletz=3.025; % z-offset of outlet
OutWall=1; % 1=Outlet is on same side as supply, 2=opposite wall.

Contz=1.5; % z-position of fronts of load banks (LBs)
Servx=0.8*3+0.8/4; % x-offset of server (m). Approximation for rack 4, based on server occupying half of width of rack.
Servy=0.5; % y-offset of server (m)
HServ=0.3; % height of server (m)
WServ=0.4; % width of server (m)
LServ=0.8; % length of server (m)
EServ=6.761; % power consumption of server (consistent with 6 heaters on in total
VServ=0.1969; % flow rate through server (this is only used to initialise) (m^3/s)
RackDepth=1.1; % Depth of rack
RackHeight=2.1; % Height of rack
ServerPts=(HServ/Dy)*(WServ/Dx);  % number of nodes at server inlet (used later to determine the inlet velocity required to achieve the specified volumetric flow rate)
EmptySlot=0; % Empty slot above server? 1=yes.
SlotPts=1*WServ/Dx;
BypassPts=0; % Number of nodes through which bypass is allowed
% Initialise supply flow rate or pressure drop, as required
if EquilCond==1
    p_CA=((VIn/VServ)-1)*50;
else
    VIn=(1+p_CA/50)*VServ;
end
VInStore=VIn;

%Now calculate server fan power consumption at local dp=0
C_D=38289; % discharge coefficient of v20z server, from Brady (2016 p203) (Pa/m^3/s)
P_required=C_D*(VServ)^2; % Pressure to be developed by internal server fans

if HInlet<2*Dy || DInlet<2*Dz
    'ERROR - Inlet too small in comparison to grid spacing'
elseif HOutlet<2*Dy || DOutlet<2*Dz
    'ERROR - Outlet too small in comparison to grid spacing'
elseif Inlety<Dy || Inletz<Dz || Inlety+HInlet>=H || Inletz+DInlet>=D
    'ERROR - Inlet does not fit within domain'
elseif Outletz<Dz || Outlety+HOutlet>=H || Outletz+DOutlet>=D
    'ERROR - Outlet does not fit within domain'
end

if Dx==Dy && Dy==Dz
else
    'ERROR - Dx/=Dy or Dy/=Dz, hence continuity equation is invalid'
end

%% Solution parameters etc.
NIterations=50000; % Max number of iterations
WSOR=1.2; % relaxation factor for SOR
BuoyancyOn=0; % (set to 1 to include buoyancy, 0 to ignore buoyancy)
OutNeumann=1; % Neumann boundary condition at outlet? 1 for yes, 0 for zero velocity gradient BC, 2 for variable velocity but fixed flow rate.
EnvHeatLoad=0.0015; % Heat load on external walls of cold aisle (kW/m^2)

%% Initialise variables
Phi=zeros(nx+2,ny+2,nz+2); % velocity potential, with ghost cells at boundary
[p_c p_h p_cCount p_hCount]=deal(0);
p=p_0*ones(nx+2,ny+2,nz+2);
T=(T_0+(EnvHeatLoad*(L*H+H*2*Contz+Contz*L)/(c_p*rho_0*VIn)))*ones(nx+2,ny+2,nz+2); % temperature

BCs=zeros(nx+2,ny+2,nz+2); % Boundary conditions
kcont=round(Contz/Dz+1); % 'k' correlating with containment partition/front of rack
%% Store codes for boundary condition positions
InletPts=0;
OutletPts=0;
for j=1:ny+2
    for k=1:nz+2
        if ((j*Dy>(Inlety-0.5*HInlet) && j*Dy<(Inlety+0.5*HInlet)) && (k*Dz>(Inletz-0.5*DInlet) && k*Dz<(Inletz+0.5*DInlet))) && ((j*Dy>Outlety && j*Dy<(Outlety+HOutlet)) && (k*Dz>Outletz && k*Dz<(Outletz+DOutlet)))
            'ERROR - Inlet & Outlet overlap';
            return
        elseif ((j-1.5)*Dy>Inlety && (j-1.5)*Dy<(Inlety+HInlet)) && ((k-1.5)*Dz>Inletz && (k-1.5)*Dz<(Inletz+DInlet))
            BCs(1,j,k)=2; % 2 denotes inlet
            InletPts=InletPts+1; % This line counts the number of cells representing the inlet, which is used later to determine the required Phi values at the inlet
            BCs(nx+2,j,k)=1; % 1 denotes solid wall (opposite wall)
        elseif ((j-1.5)*Dy>Outlety && (j-1.5)*Dy<(Outlety+HOutlet)) && ((k-1.5)*Dz>Outletz && (k-1.5)*Dz<(Outletz+DOutlet))
            if OutWall==1
                BCs(1,j,k)=3; % 3 denotes outlet
                BCs(nx+2,j,k)=1; % 1 denotes solid wall (opposite wall)
            else
                BCs(nx+2,j,k)=3; % 3 denotes outlet
                BCs(1,j,k)=1; % 1 denotes solid wall (opposite wall)
            end
            OutletPts=OutletPts+1; % This line counts the number of cells representing the outlet, which is used later to determine the required Phi values at the outlet
        else
            BCs(1,j,k)=1; % 1 denotes solid wall
            BCs(nx+2,j,k)=1;
        end
        
        % rack walls and roofs
        if k>kcont && k<=kcont+RackDepth/Dz
            if ((j-1.5)*Dy>RackHeight && (j-1.5)*Dy<=RackHeight+Dy)
                for i=2:nx+1
                    BCs(i,j,k)=9; % 9=rack roof
                end
            end
        end
    end
end

for i=1:nx+2
    for k=1:nz+2
        BCs(i,1,k)=1; % 1 denotes solid wall
        BCs(i,ny+2,k)=1;
    end
end

ContFlow=0; % Resets record of flow rate through the containment partition
Vtemp=0;
for i=2:nx+1
    for j=2:ny+1
        for k=2:nz+1
            if k==kcont
                if (i>=ceil(1.5+Servx/Dx) && i<=ceil(1.5+Servx/Dx+round(WServ/Dx)-1)) && (j>=round(0.5*Dy+Servy/Dy) && j<=round(0.5*Dy+Servy/Dy)+round(HServ/Dy)-1) % within area of a server
                    BCs(i,j,k)=5; % 5 denotes server inlet
                    if EmptySlot==1 && j+1>round(0.5*Dy+Servy/Dy)+round(HServ/Dy)-1 % i.e. if there is an empty slot and the next node up is not within server
                        BCs(i,j+1:j+2,k)=11; % denotes empty slot
                    end
                    for c=1:floor(LServ/Dz)
                        BCs(i,j,k+c)=6; % Inside server
                    end
                    BCs(i,j,k+c+1)=7; % Server outlet
                else % not within area of server
                    if BCs(i,j,k)==11 % Empty slot
                    else
                        BCs(i,j,k)=4; % 4 denotes rack front/aisle containment
                    end
                end
                ContFlow=ContFlow+Vtemp*Dx*Dy; % Total bypass
                ContFlow=ContFlow+Vtemp*Dx*Dy; % Total flow from cold to hot aisle
            end
        end
    end
end    
for i1=1:N_racks+1
    for i2=-floor(0.5*round(0.4/CellSide))+1:ceil(0.5*round(0.4/CellSide))
        for i3=2:floor((2.1/Dy)+1.5)
            if round(1+(i1-1)*0.8/CellSide+i2)>1 && round(1+(i1-1)*0.8/CellSide+i2)<nx+2
                if BCs(round(1+(i1-1)*0.8/CellSide+i2),i3,kcont)==4 % This makes sure we're not over-riding a server inlet
                    BCs(round(1+(i1-1)*0.8/CellSide+i2),i3,kcont)=8; % leakage path
                    BypassPts=BypassPts+1;
                end
            end
        end
    end
end
for i1=1:N_racks-1
    BCs(round(i1*0.8/CellSide)+1,2:round(RackHeight/CellSide+1.5),kcont+1:round(kcont+RackDepth/CellSide))=10; % 10=rack side
end

istore=nx+2; % These are used in the results plotting program
jstore=ny+2;
kstore=nz+2;

%% Convergence parameters etc.
iteration=2; % Iteration 1 being the initial conditions
PhiRes=1;
PhiConvergedIt=0;
PhiResTarget=0.000001;
pRes=1;
pConvergedIt=0;
pResTarget=0.004;
TRes=1;
TConvergedIt=0;
MaxTChange=5*ones(NIterations,1);
MaxTChangeTarget=0.05;

%% Begin iteration
while iteration<=NIterations && TConvergedIt==0
%     if iteration/25==floor(iteration/25) % A monitor for how far the
%     simulation has progessed
%         iteration
%         PhiRes
%     end
    ContFlow=0;
    ContFlowStore(iteration)=0;
    if (PhiRes>PhiResTarget || pRes>pResTarget) && PhiConvergedIt<2
        
        %% calculate bulk pressure difference between h&c aisles
        [p_c p_h p_cCount p_hCount]=deal(0);
        for i=2:nx+1 % Ignore ghost cells
            for j=2:ny+1
                ColdAisle=1;
                for k=2:nz+1
                    if k>kcont
                        ColdAisle=0;
                    end
                    if ColdAisle==1 && isnan(p(i,j,k,1))==0
                        p_c=p_c+p(i,j,k,1);
                        p_cCount=p_cCount+1;
                    elseif isnan(p(i,j,k,1))==0
                        p_h=p_h+p(i,j,k,1);
                        p_hCount=p_hCount+1;
                    end
                end
            end
        end
        dp=(p_c/p_cCount)-(p_h/p_hCount); % average dp across containment
        dpStore(iteration)=dp;
        
        %% Phi calculation
        % first set Phi at domain boundaries where i/j/k=1
        for j=1:ny+2 % walls in the x-plane
            for k=1:nz+2
                if BCs(1,j,k)==1 % solid wall
                    if BCs(3,j,k)==6 % If adcacent cell is adjacent to server, need to avoid Phi from this cell staying at 0 regardless of changes to Phi in adjacent cells
                        Phi(1,j,k,2)=Phi(1,j-1,k,2);
                    else
                        Phi(1,j,k,2)=Phi(2,j,k,2-1);
                    end
                elseif BCs(1,j,k)==2 % inlet
                    Phi(1,j,k,2)=Phi(2,j,k,2-1)+(VIn/(InletPts*Dy*Dz))*Dx; % Just gives constant velocity across the inlet.
                elseif BCs(1,j,k)==3 % outlet
                    if OutNeumann==1 % Neumann BC at outlet
                        Phi(1,j,k,2)=Phi(2,j,k,2-1)-(VIn/(OutletPts*Dy*Dz))*Dx; % Just gives constant velocity across the outlet.
                    elseif OutNeumann==0 % Zero velocity gradient
                        Phi(1,j,k,2)=Phi(2,j,k,2-1)+(Phi(3,j,k,2-1)-Phi(2,j,k,2-1)); % Free boundary, velocity gradient=0
                    else % Variable velocity, fixed flow rate
                        if PhiOutAvDone==0
                            PhiOutAv=(sum(sum(sum(sum(Phi(2,j:j+floor(HOutlet/Dy),k:k+floor(DOutlet/Dz),2))))))/OutletPts;
                            PhiOutAvDone=1;
                        end
                        Phi(1,j,k,2)=PhiOutAv-(VIn/(OutletPts*Dy*Dz))*Dx; % Should give the right flow rate across the outlet, but allows velocity to fluctuate                        
                    end
                end
                if iteration==2 % only set Phi at i/j/k=nx/ny/nz, since in later iterations this calculation should use the most up to date neighbouring Phi values
                    if BCs(nx+2,j,k)==1
                        Phi(nx+2,j,k,2)=Phi(nx+1,j,k,2-1);
                    elseif BCs(nx+2,j,k)==2
                        Phi(nx+2,j,k,2)=Phi(nx+1,j,k,2-1)+(VIn/(InletPts*Dy*Dz))*Dx; % Just gives constant velocity across the inlet.
                    elseif BCs(nx+2,j,k)==3
                        Phi(nx+2,j,k,2)=Phi(nx+1,j,k,2-1)-(VIn/(OutletPts*Dy*Dz))*Dx; % Just gives constant velocity across the outlet.
                    end
                end
            end
        end
        for i=1:nx+2 % walls in the y-plane
            for k=1:nz+2
                if BCs(i,1,k)==1 % solid wall
                    if BCs(i,3,k)==6
                        Phi(i,1,k,2)=Phi(i-1,1,k,2);
                    else
                        Phi(i,1,k,2)=Phi(i,2,k,2-1);
                    end
                elseif BCs(i,1,k)==2 % inlet
                    Phi(i,1,k,2)=Phi(i,2,k,2-1)+(VIn/(InletPts*Dx*Dz))*Dy; % Just gives constant velocity across the inlet.
                elseif BCs(i,1,k)==3 % outlet
                    if OutNeumann==1 % Neumann BC at outlet
                        Phi(i,1,k,2)=Phi(i,2,k,2-1)-(VIn/(OutletPts*Dx*Dz))*Dy; % Just gives constant velocity across the outlet.
                    elseif OutNeumann==0 % Zero velocity gradient
                        Phi(i,1,k,2)=Phi(i,2,k,2-1)+(Phi(i,3,k,2-1)-Phi(i,2,k,2-1)); % Free boundary, velocity gradient=0
                    else % Variable velocity, fixed flow rate
                        if PhiOutAvDone==0
                            PhiOutAv=(sum(sum(sum(sum(Phi(i:i+floor(WOutlet/Dy),2,k:k+floor(DOutlet/Dz),2))))))/OutletPts;
                            PhiOutAvDone=1;
                        end
                        Phi(i,1,k,2)=PhiOutAv-(VIn/(OutletPts*Dx*Dz))*Dy; % Should give the right flow rate across the outlet, but allows velocity to fluctuate
                    end
                end
                if iteration==2
                    if BCs(i,ny+2,k)==1
                        Phi(i,ny+2,k,2)=Phi(i,ny+1,k,2-1);
                    elseif BCs(i,ny+2,k)==2
                        Phi(i,ny+2,k,2)=Phi(i,ny+1,k,2-1)+(VIn/(InletPts*Dx*Dz))*Dy; % Just gives constant velocity across the inlet.
                    elseif BCs(i,ny+2,k)==3
                        Phi(i,ny+2,k,2)=Phi(i,ny+1,k,2-1)-(VIn/(OutletPts*Dx*Dz))*Dy; % Just gives constant velocity across the outlet.
                    end
                end
            end
        end
        for i=1:nx+2 % walls in the z-plane
            for j=1:ny+2
                if BCs(i,j,1)==1 % solid wall
                    if BCs(i,j,3)==6
                        Phi(i,j,1,2)=Phi(i,j-1,1,2);
                    else
                        Phi(i,j,1,2)=Phi(i,j,2,2-1);
                    end
                elseif BCs(i,j,1)==2 % inlet
                    Phi(i,j,1,2)=Phi(i,j,2,2-1)+(VIn/(InletPts*Dx*Dy))*Dz; % Just gives constant velocity across the inlet.
                elseif BCs(i,j,1)==3 % outlet
                    if OutNeumann==1 % Neumann BC at outlet
                        Phi(i,j,1,2)=Phi(i,j,2,2-1)-(VIn/(OutletPts*Dx*Dy))*Dz; % Just gives constant velocity across the inlet.
                    elseif OutNeumann==0 % Zero velocity gradient
                        Phi(i,j,1,2)=Phi(i,j,2,2-1)+(Phi(i,j,3,2-1)-Phi(i,j,2,2-1)); % Free boundary, velocity gradient=0
                    else % Variable velocity across outlet, fixed flow rate
                        if PhiOutAvDone==0
                            PhiOutAv=(sum(sum(sum(sum(Phi(i:i+floor(WOutlet/Dy),j:j+floor(HOutlet/Dy),2,2))))))/OutletPts;
                            PhiOutAvDone=1;
                        end
                        Phi(i,j,1,2)=PhiOutAv-(VIn/(OutletPts*Dx*Dy))*Dz; % Should give the right flow rate across the outlet, but allows velocity to fluctuate
                    end
                end
                if iteration==2
                    if BCs(i,j,nz+2)==1
                        Phi(i,j,nz+2,2)=Phi(i,j,nz+1,2-1);
                    elseif BCs(i,j,nz+2)==2
                        Phi(i,j,nz+2,2)=Phi(i,j,nz+1,2-1)+(VIn/(InletPts*Dx*Dy))*Dz; % Just gives constant velocity across the inlet.
                    elseif BCs(i,j,nz+2)==3
                        Phi(i,j,nz+2,2)=Phi(i,j,nz+1,2-1)-(VIn/(OutletPts*Dx*Dy))*Dz; % Just gives constant velocity across the outlet.
                    end
                end
            end
        end

        for i=1:nx+2
            for j=1:ny+2
                for k=2:nz+1
                    if BCs(i,j,k)==4 || BCs(i,j,k)==5 % containment
                        kstart=k;
                        break
                    end
                end
            end
        end
        for k=kstart:-1:2 % Now carry out Phi calculation for the cold aisle
            for j=2:ny+1
                for i=2:nx+1
                    if k>2
                        if k==kstart
                            Phi(i,j,k,2)=WSOR*((1/6)*(Phi(i-1,j,k,2)+Phi(i+1,j,k,2-1)+Phi(i,j-1,k,2)+Phi(i,j+1,k,2-1)+Phi(i,j,k-1,2-1)+Phi(i,j,k+1,2-1)))+(1-WSOR)*Phi(i,j,k,2-1);
                        else
                            Phi(i,j,k,2)=WSOR*((1/6)*(Phi(i-1,j,k,2)+Phi(i+1,j,k,2-1)+Phi(i,j-1,k,2)+Phi(i,j+1,k,2-1)+Phi(i,j,k-1,2-1)+Phi(i,j,k+1,2)))+(1-WSOR)*Phi(i,j,k,2-1);
                        end
                    else % Phi is available at (i,j,1) for the current iteration (i.e. at the boundary)
                        Phi(i,j,k,2)=WSOR*((1/6)*(Phi(i-1,j,k,2)+Phi(i+1,j,k,2-1)+Phi(i,j-1,k,2)+Phi(i,j+1,k,2-1)+Phi(i,j,k-1,2)+Phi(i,j,k+1,2)))+(1-WSOR)*Phi(i,j,k,2-1);
                    end
                end
            end
        end
        LeakCount=0;
        ContFlowStore(iteration)=0;
        SlotFlowStore(iteration)=0;
        ServerFanPowerTemp=0;
        for i=2:nx+1
            for j=2:ny+1
                if BCs(i,j,kstart)==5 % if server inlet
                    for m=1:nz+2
                        if BCs(i,j,m)==7 % server outlet
                            break
                        end
                    end
                    dptemp=p(i,j,kstart-1,1)-p(i,j,m+1,1); % Local pressure drop across server
                    dptempstoreserv(i,j)=dptemp;
                    Phi(i,j,kstart+1,PhiIt)=Phi(i,j,kstart,PhiIt)-Dz*(VServ/(ServerPts*Dx*Dy));
                                        
                elseif BCs(i,j,kstart)==4 % containment, zero flow
                    Phi(i,j,kstart+1,PhiIt)=Phi(i,j,kstart,PhiIt);
                    Phi(i,j,kstart+2,PhiIt)=Phi(i,j,kstart+3,PhiIt-1);
                else % leakage path, flows set according to equations derived in Tatchell-Evans (2017, Chapter 2)
                    if BCs(i,j,kstart)==11 % Empty slot
                        dptemp=p(i,j,kstart,1)-p(i,j,kstart+1,1); % Local pressure drop across partition
                        dptempstoreSlot(i,j)=dptemp;
                        if dptemp<0
                            SlotFlow=-((-dptemp/486)^0.5-(-dptemp/1502)^0.5);
                        else
                            SlotFlow=((-7.587+(7.587^2+4*441.88*dptemp)^0.5)/(2*441.88))-((-28.741+(28.741^2+4*1121.1*dptemp)^0.5)/(2*1121.1));
                        end
                        SlotFlowStore(iteration)=SlotFlowStore(iteration)+SlotFlow/SlotPts;
                        Phi(i,j,kstart+1,PhiIt)=Phi(i,j,kstart,PhiIt)-((SlotFlow/SlotPts)/(Dx*Dy))*Dz;
                        Phi(i,j,kstart+2,PhiIt)=Phi(i,j,kstart+3,PhiIt-1)+((SlotFlow/SlotPts)/(Dx*Dy))*Dz;
                    else % Normal leakage path
                        dptemp=p(i,j,kstart,1)-p(i,j,kstart+1,1); % Local pressure drop across partition
                        dptempstore(i,j)=dptemp;
                        if dptemp<0
                            % Low bypass (Equation gives bypass per rack, including one open server interface per rack)
                            ContFlow=-((-80.038+(80.038^2+4*452.89*-dptemp)^0.5)/(2*452.89))+0.75*((-56.592+(56.592^2+4*1289*-dptemp)^0.5)/(2*1289));
                            % High bypass (original equation is for 4 racks, hence 0.25 factor added)
                            % ContFlow=-0.25*(N_racks/4)*((-3.689+(3.689^2+4*45.855*-dptemp)^0.5)/(2*45.855));
                        else
                            % Low bypass (Equation gives bypass per rack, including one open server interface per rack)
                            ContFlow=((-80.038+(80.038^2+4*452.89*dptemp)^0.5)/(2*452.89))-0.75*((-56.592+(56.592^2+4*1289*dptemp)^0.5)/(2*1289));
                            % High bypass (original equation is for 4 racks, hence 0.25 factor added)
                            % ContFlow=0.25*((-3.689+(3.689^2+4*45.855*dptemp)^0.5)/(2*45.855));
                        end
                        ContFlow=ContFlow*N_racks; % Scale up to full aisle
                        ContFlowStore(iteration)=ContFlowStore(iteration)+ContFlow/BypassPts;
                        Phi(i,j,kstart+1,PhiIt)=Phi(i,j,kstart,PhiIt)-((ContFlow/BypassPts)/(Dx*Dy))*Dz;
                    end
                end
            end
        end
        
        ServerFlowResult=0; % Reset flow through server
        for k=kstart+2:nz+1 % Hot aisle
            for j=2:ny+1
                for i=2:nx+1
                    if BCs(i,j,k)==6 % Inside server
                        if BCs(i-1,j,k)~=6
                            Phi(i,j,k,PhiIt)=Phi(i-1,j,k,PhiIt);
                        elseif BCs(i+1,j,k)~=6
                            Phi(i,j,k,PhiIt)=Phi(i+1,j,k,PhiIt-1);
                        elseif BCs(i,j-1,k)~=6
                            Phi(i,j,k,PhiIt)=Phi(i,j-1,k,PhiIt);
                        elseif BCs(i,j+1,k)~=6
                            Phi(i,j,k,PhiIt)=Phi(i,j+1,k,PhiIt-1);
                        else
                            Phi(i,j,k,PhiIt)=NaN;
                        end
                        
                    elseif BCs(i,j,k)==7 % Server outlet
                        Phi(i,j,k,PhiIt)=Phi(i,j,k+1,PhiIt-1)+(Phi(i,j,kstart,PhiIt)-Phi(i,j,kstart+1,PhiIt)); % Ensures flow into server=flow out of server                        
                    elseif BCs(i,j,k)==9 % rack roof
                        Phi(i,j,k,PhiIt)=Phi(i,j-1,k,PhiIt);
                    elseif BCs(i,j,k)==10 % rack side
                        Phi(i,j,k,PhiIt)=Phi(i-1,j,k,PhiIt);                        
                    else % in open space i.e. no BCs apply
                        Phi(i,j,k,PhiIt)=WSOR*((1/6)*(Phi(i-1,j,k,PhiIt)+Phi(i+1,j,k,PhiIt-1)+Phi(i,j-1,k,PhiIt)+Phi(i,j+1,k,PhiIt-1)+Phi(i,j,k-1,PhiIt)+Phi(i,j,k+1,PhiIt-1)))+(1-WSOR)*Phi(i,j,k,PhiIt-1);
                    end
                end
            end
        end

        % Now domain boundaries
        PhiOutAvDone=0;
        for j=1:ny+2
            for k=1:nz+2
                if BCs(nx+2,j,k)==1
                    if BCs(nx,j,k)==6 % If adcacent cell is adjacent to server, need to avoid Phi from this cell staying at 0 regardless of changes to Phi in adjacent cells
                        Phi(nx+2,j,k,PhiIt)=Phi(nx+2,j-1,k,PhiIt);
                    else
                        Phi(nx+2,j,k,PhiIt)=Phi(nx+1,j,k,PhiIt);
                    end
                elseif BCs(nx+2,j,k)==2
                    Phi(nx+2,j,k,PhiIt)=Phi(nx+1,j,k,PhiIt)+(VIn/(InletPts*Dy*Dz))*Dx; % Just gives constant velocity across the inlet.
                elseif BCs(nx+2,j,k)==3
                    if OutNeumann==1 % Neumann BC or zero velocity gradient at outlet
                        Phi(nx+2,j,k,PhiIt)=Phi(nx+1,j,k,PhiIt)-(VIn/(OutletPts*Dy*Dz))*Dx; % Just gives constant velocity across the inlet.
                    elseif OutNeumann==0 % Zero velocity gradient
                        Phi(nx+2,j,k,PhiIt)=Phi(nx+1,j,k,PhiIt)-(Phi(nx,j,k,PhiIt)-Phi(nx+1,j,k,PhiIt)); % Free boundary, velocity gradient=0
                    else % Variable velocity, fixed flow rate
                        if PhiOutAvDone==0
                            PhiOutAv=(sum(sum(sum(sum(Phi(nx+1,j:j+floor(HOutlet/Dy),k:k+floor(DOutlet/Dz),PhiIt))))))/OutletPts;
                            PhiOutAvDone=1;
                        end
                        Phi(nx+2,j,k,PhiIt)=PhiOutAv-(VIn/(OutletPts*Dy*Dz))*Dx; % Should give the right flow rate across the outlet, but allows it to fluctuate                        
                    end
                end
            end
        end
        for i=1:nx+2
            for k=1:nz+2
                if BCs(i,ny+2,k)==1 % solid wall
                    if BCs(i,ny,k)==6
                        Phi(i,ny+2,k,PhiIt)=Phi(i-1,ny+2,k,PhiIt);
                    else
                        Phi(i,ny+2,k,PhiIt)=Phi(i,ny+1,k,PhiIt);
                    end
                elseif BCs(i,ny+2,k)==2 % inlet
                    Phi(i,ny+2,k,PhiIt)=Phi(i,ny+1,k,PhiIt)+(VIn/(InletPts*Dx*Dz))*Dy; % Just gives constant velocity across the inlet.
                elseif BCs(i,ny+2,k)==3 % outlet
                    if OutNeumann==1 % Neumann BC or zero velocity gradient at outlet
                        Phi(i,ny+2,k,PhiIt)=Phi(i,ny+1,k,PhiIt)-(VIn/(OutletPts*Dx*Dz))*Dy; % Just gives constant velocity across the outlet.
                    elseif OutNeumann==0 % Zero velocity gradient
                        Phi(i,ny+2,k,PhiIt)=Phi(i,ny+1,k,PhiIt)-(Phi(i,ny,k,PhiIt)-Phi(i,ny+1,k,PhiIt)); % Free boundary, velocity gradient=0
                    else % Variable velocity, fixed flow rate
                        if PhiOutAvDone==0
                            PhiOutAv=(sum(sum(sum(sum(Phi(i:i+floor(WOutlet/Dy),ny+1,k:k+floor(DOutlet/Dz),PhiIt))))))/OutletPts;
                            PhiOutAvDone=1;
                        end
                        Phi(i,ny+2,k,PhiIt)=PhiOutAv-(VIn/(OutletPts*Dx*Dz))*Dy; % Should give the right flow rate across the outlet, but allows it to fluctuate
                    end
                end
            end
        end
        for i=1:nx+2
            for j=1:ny+2
                if BCs(i,j,nz+2)==1 % solid wall
                    if BCs(i,j,nz)==6
                        Phi(i,j,nz+2,PhiIt)=Phi(i-1,j,nz+2,PhiIt);
                    else
                        Phi(i,j,nz+2,PhiIt)=Phi(i,j,nz+1,PhiIt);
                    end
                elseif BCs(i,j,nz+2)==2 % inlet
                    Phi(i,j,nz+2,PhiIt)=Phi(i,j,nz+1,PhiIt)+(VIn/(InletPts*Dx*Dy))*Dz; % Just gives constant velocity across the inlet.
                elseif BCs(i,j,nz+2)==3 % outlet
                    if OutNeumann==1 % Neumann BC or zero velocity gradient at outlet
                        Phi(i,j,nz+2,PhiIt)=Phi(i,j,nz+1,PhiIt)-(VIn/(OutletPts*Dx*Dy))*Dz; % Just gives constant velocity across the outlet.
                    elseif OutNeumann==0 % Zero velocity gradient
                        Phi(i,j,nz+2,PhiIt)=Phi(i,j,nz+1,PhiIt)-(Phi(i,j,nz,PhiIt)-Phi(i,j,nz+1,PhiIt)); % Free boundary, velocity gradient=0
                    else % Variable velocity across outlet, fixed flow rate
                        if PhiOutAvDone==0
                            PhiOutAv=(sum(sum(sum(sum(Phi(i:i+floor(WOutlet/Dy),j:j+floor(HOutlet/Dy),nz+1,PhiIt))))))/OutletPts;
                            PhiOutAvDone=1;
                        end
                        Phi(i,j,nz+2,PhiIt)=PhiOutAv-(VIn/(OutletPts*Dx*Dy))*Dz; % Should give the right flow rate across the outlet, but allows it to fluctuate
                    end
                end
            end
        end

        %% Calculate velocities
[ustore vstore wstore]=deal(0);
        for i=1:nx+2 % The nodes of the velocity field correspond with the cell faces
            for j=1:ny+2
                for k=1:nz+2
                    if i<nx+2
                        u(i,j,k)=(Phi(i,j,k,PhiIt)-Phi(i+1,j,k,PhiIt))/Dx;
                        if isnan(u(i,j,k))==1
                            u(i,j,k)=0;
                        end
                    end
                    if j<ny+2
                        v(i,j,k)=(Phi(i,j,k,PhiIt)-Phi(i,j+1,k,PhiIt))/Dy;
                        if isnan(v(i,j,k))==1
                            v(i,j,k)=0;
                        end
                    end
                    if k<nz+2
                        if k==kstart+1
                            w(i,j,k)=w(i,j,k-1); % Work around for flow through partition
                        else
                            w(i,j,k)=(Phi(i,j,k,PhiIt)-Phi(i,j,k+1,PhiIt))/Dz;
                            if isnan(w(i,j,k))==1
                                w(i,j,k)=0;
                                wnanStore(i,j,k)=1;
                            end
                        end
                    end
                end
            end
        end

    else % if Phi field has converged, set Phi and other relevant variables at same values as previous iteration
        ContFlowStore(iteration)=ContFlowStore(iteration-1);
        SlotFlowStore(iteration)=SlotFlowStore(iteration-1);
        Phi(1:nx+2,1:ny+2,1:nz+2,PhiIt)=Phi(1:nx+2,1:ny+2,1:nz+2,PhiIt-1);
           
    %% advection
        maxV=(max(max(max(u))));
        if (max(max(max(v))))>maxV
            maxV=(max(max(max(v))));
        end
        if (max(max(max(w))))>maxV
            maxV=(max(max(max(w))));
        end
        if size(T,4)<iteration
            T(1:nx+2,1:ny+2,1:nz+2,iteration)=T(1:nx+2,1:ny+2,1:nz+2,iteration-1);
        end
        
        T_HAcurrent=mean(mean(mean(mean(T(1:nx+2,1:ny+2,kstart+1+floor(RackDepth/Dz):nz+2,iteration))))); % calculate mean HA temperature for use in calculation of heat loss through the walls
        T_CAcurrent=mean(mean(mean(mean(T(1:nx+2,1:ny+2,1:kstart+floor(RackDepth/Dz),iteration))))); % calculate mean CA temperature for use in calculation of heat loss through the walls
        for i=2:nx+1 % T nodes correspond to Phi nodes, no need to calculate T in the ghost cells
            Rack=floor((((i-1.5)*Dx)+1)/0.8); % Identifies rack number
            for j=2:ny+1
                for k=2:nz+1
                    if BCs(i,j,k)==2 % Inlet to test data centre
                        T(i,j,k,iteration)=T(i,j,k,iteration-1); % maintains temperature of inlet air
                    elseif BCs(i,j,k)==6
                        T(i,j,k,iteration)=T_0;
                    elseif BCs(i,j,k)==7 % Server outlet
                        T(i,j,k,iteration)=T(i,j,kstart,iteration)+(EServ/ServerPts)/((VServ/ServerPts)*rho_0*c_p); % First term in addition is the temperature at the inlet cell opposite to this outlet cell.

                    else
                        Vcell=0;
                        u0=[u(i-1,j,k) -u(i,j,k)];
                        v0=[v(i,j-1,k) -v(i,j,k)];
                        w0=[w(i,j,k-1) -w(i,j,k)];
                        if isnan(u0+v0+w0)==1
                            'u0 etc nan'
                            pause(5)
                        end
                        if u(i-1,j,k)+u(i,j,k)+v(i,j-1,k)+v(i,j,k)+w(i,j,k-1)+w(i,j,k)==0 || isnan(u(i-1,j,k)+u(i,j,k)+v(i,j-1,k)+v(i,j,k)+w(i,j,k-1)+w(i,j,k))==1
                            T(i,j,k,iteration)=T(i,j,k,iteration-1);
                        else
                            Vcell=sum(u0)+sum(v0)+sum(w0); % This represents the excess flow into the cell, should be 0
                            HeatEnt=((u0(1)>0)*u0(1)*T(i-1,j,k,iteration)+(u0(2)>0)*u0(2)*T(i+1,j,k,iteration-1)+(v0(1)>0)*v0(1)*T(i,j-1,k,iteration)+(v0(2)>0)*v0(2)*T(i,j+1,k,iteration-1)+(w0(1)>0)*w0(1)*T(i,j,k-1,iteration)+(w0(2)>0)*w0(2)*T(i,j,k+1,iteration-1)); % Heat entering the cell through convection
                            HeatLeave=-((u0(1)<0)*u0(1)+(u0(2)<0)*u0(2)+(v0(1)<0)*v0(1)+(v0(2)<0)*v0(2)+(w0(1)<0)*w0(1)+(w0(2)<0)*w0(2));
                            T(i,j,k,iteration)=(HeatEnt-((k>kstart+1)*U_cube*(T_HAcurrent-T_0)/(rho_0*c_p*(Dx^2))))/HeatLeave; % applying heat loss through walls
                            i1=0;
                            Tadj=T(i,j,k,iteration-1);
                            if u(i-1,j,k)>0
                                i1=i1+1;
                                Tadj(i1)=T(i-1,j,k,iteration);
                            end
                            if u(i,j,k)<0
                                i1=i1+1;
                                Tadj(i1)=T(i+1,j,k,iteration);
                            end
                            if v(i,j-1,k)>0
                                i1=i1+1;
                                Tadj(i1)=T(i,j-1,k,iteration);
                            end
                            if v(i,j,k)<0
                                i1=i1+1;
                                Tadj(i1)=T(i,j+1,k,iteration);
                            end
                            if w(i,j,k-1)>0
                                i1=i1+1;
                                Tadj(i1)=T(i,j,k-1,iteration);
                            end
                            if w(i,j,k)<0
                                i1=i1+1;
                                Tadj(i1)=T(i,j,k+1,iteration);
                            end
                            if T(i,j,k,iteration)<min(Tadj)
                                Tamended(i,j,k)=min(Tadj)-T(i,j,k,iteration);
                                T(i,j,k,iteration)=min(Tadj);
                            elseif T(i,j,k,iteration)>max(Tadj)
                                Tamended(i,j,k)=max(Tadj)-T(i,j,k,iteration);
                                T(i,j,k,iteration)=max(Tadj);
                            end
                            if T(i,j,k,iteration)<T_0 % Check for errors causing temperatures below inlet temperature
                                Tamended(i,j,k)=T(i,j,k,iteration);
                                T(i,j,k,iteration)=T_0;
                            end
                        end
                    end
                end
            end
        end
        T(1:nx+2,1:ny+2,1,2)=T(1:nx+2,1:ny+2,2,2); % Symmetry about centre of cold aisle
        T(1:nx+2,1:ny+2,nz+2,2)=T(1:nx+2,1:ny+2,nz+1,2); % Symmetry about centre of hot aisle
        if isreal(u)==0 || isreal(v)==0 || isreal(w)==0 || isreal(p)==0
            'complex'
            pause(5)
            notcomplex=2;
        else
            notcomplex=1;
        end

        % Calculate T residual
        if TConvergedIt==0
            if MaxTChange(iteration-1)>MaxTChangeTarget
                TRes=0;
                TSum=0;
                for i=1:nx
                    for j=1:ny
                        for k=1:nz
                            if isnan(T(i,j,k,iteration))==0
                                TRes=TRes+abs(T(i,j,k,iteration)-T(i,j,k,iteration-1));
                                TSum=TSum+abs(T(i,j,k,iteration));
                                TChange(i,j,k)=abs(T(i,j,k,iteration)-T(i,j,k,iteration-1));
                            end
                        end
                    end
                end
                MaxTChange(iteration)=max(TChange(:));
                TRes=TRes/TSum;
            else
                TConvergedIt=iteration;
            end
        end
        
        % Now add buoyancy
        if BuoyancyOn==1
            if iteration==PhiConvergedIt+1 || iteration==pConvergedIt+1
                uNB=u; % store velocity fields prior to the addition of buoyancy
                vNB=v;
                wNB=w;
                'Buoyancy kicked in'
            end
            [T_H T_H0 T_H1 T_H2 T_H3 T_H4 T_HCount T_H0Count T_H1Count T_H2Count T_H3Count T_H4Count]=deal(0);
            [v_H0 v_H1 v_H2 v_H3 v_H4 v_H0Count v_H1Count v_H2Count v_H3Count v_H4Count]=deal(0);
            H_buoy=Servy+HServ/2; % This is the position of the Rankine Vortex in the y-dimension (set to half the rack height in Lettieri et al 2013).
            Buoyx=L*0.75-0.4; % Centre of vortex along length of aisle
            V_s=2; % Vortex strength (Lettieri et al used values up to 4, noting 2 made sense since this is multiplied by H_buoy, which they had set to half the rack height, to give a characteristic length scale equal to the rack height).
            R=0.8; % This is the radius parameter for the Rankine vortex, and determines where the behaviour flips from forced vortex to potential vortex
            % Calculate average v and T in the hot aisle
            for i=2:nx+1
                Rack=floor((((i-1.5)*Dx)+1)/0.8); % Identifies rack number
                Rack=Rack-(Rack>N_racks);
                for j=2:ny+1
                    for k=((floor((Contz+LServ+0.5*Dz)/Dz))+2):nz+1 % Hot aisle
                        if BCs(i,j,k)==0
                            T_H=T_H+T(i,j,k,iteration);
                            T_HCount=T_HCount+1;
                            if i<0.25*nx+1
                                T_H0=T_H0+T(i,j,k,iteration); % This refers to the region behind rack 1
                                T_H0Count=T_H0Count+1;
                                v_H0=v_H0+vNB(i,j-1,k-1);
                            elseif i<0.5*nx+1
                                T_H1=T_H1+T(i,j,k,iteration);
                                T_H1Count=T_H1Count+1;
                                v_H1=v_H1+vNB(i,j-1,k-1);
                            elseif i<0.75*nx+1
                                T_H2=T_H2+T(i,j,k,iteration);
                                T_H2Count=T_H2Count+1;
                                v_H2=v_H2+vNB(i,j-1,k-1);
                            else
                                T_H3=T_H3+T(i,j,k,iteration);
                                T_H3Count=T_H3Count+1;
                                v_H3=v_H3+vNB(i,j-1,k-1);
                            end
                        end
                    end
                end
            end
            if isreal(v_H1)==0
                'complex1'
                pause(5)
                notcomplex=2;
            else
                notcomplex=1;
            end
            T_H=T_H/T_HCount; % Average temperature in hot aisle
            T_H0=T_H0/T_H0Count; % Av temp in region behind rack 1
            T_H1=T_H1/T_H1Count;
            T_H2=T_H2/T_H2Count;
            T_H3=T_H3/T_H3Count;
            v_H0=v_H0/T_H0Count; % Average v in region behind rack 1 without buoyancy
            v_H1=v_H1/T_H1Count;
            v_H2=v_H2/T_H2Count;
            v_H3=v_H3/T_H3Count;
            v_B0=((v_H0+2*9.81*(1/T_H0)*(T_H0-T_H)*H_buoy)^2)^0.5; % From Lettieri et al, 2013.  (1/T) term represents the volumetric thermal expansion coefficient (equation 4b).
            v_B1=((v_H1+2*9.81*(1/T_H1)*(T_H1-T_H)*H_buoy)^2)^0.5;
            v_B2=((v_H2+2*9.81*(1/T_H2)*(T_H2-T_H)*H_buoy)^2)^0.5;
            v_B3=((v_H3+2*9.81*(1/T_H3)*(T_H3-T_H)*H_buoy)^2)^0.5;
            k_v0=V_s*(v_B0-v_H0); % From Lettieri et al, 2013, equation 6
            k_v1=V_s*(v_B1-v_H1);
            k_v2=V_s*(v_B2-v_H2);
            k_v3=V_s*(v_B3-v_H3);
            if isreal(v_B1)==0
                'complex2'
                pause(5)
                notcomplex=2;
            else
                notcomplex=1;
            end
            % Calculate new velocities in v and x directions
            for i=2:nx+1
                for j=2:ny+1
                    for k=((floor((Contz+RackDepth+0.5*Dz)/Dz))+2):nz+1 % Hot aisle
                        if BCs(i,j,k)==7 || BCs(i,j,k)==6
                            Errortest=1;
                        elseif BCs(i,j,k-1)==7
                            Errortest2=0;
                        end
                        if BCs(i,j,k)==0 % Since the u and v fields are aligned differently, r and its components must be calculated differently.
                            Dx3u=-((i-1)*Dx-(Buoyx)); % This is the 'x' component of 'r', the distance from the centre of the vortex to the point at which the vortex's contribution to velocity is to be calculated.
                            Dy0u=-((j-1.5)*Dy-H_buoy);
                            Dx3v=-((i-1.5)*Dx-(Buoyx));
                            Dy0v=-((j-1)*Dy-H_buoy);
                            if Dx3u>0
                                if Dy0u>0
                                    theta3u=asin(Dy0u/((Dx3u^2+Dy0u^2)^0.5));
                                else
                                    theta3u=asin(Dy0u/((Dx3u^2+Dy0u^2)^0.5));
                                end
                            else
                                if Dy0u>0
                                    theta3u=0.5*pi()+(0.5*pi()-asin(Dy0u/((Dx3u^2+Dy0u^2)^0.5)));
                                else
                                    theta3u=-pi()-asin(Dy0u/((Dx3u^2+Dy0u^2)^0.5));
                                end
                            end
                            if Dx3v>0
                                if Dy0v>0
                                    theta3v=asin(Dy0v/((Dx3v^2+Dy0v^2)^0.5));
                                else
                                    theta3v=asin(Dy0v/((Dx3v^2+Dy0v^2)^0.5));
                                end
                            else
                                if Dy0u>0
                                    theta3v=0.5*pi()+(0.5*pi()-asin(Dy0v/((Dx3v^2+Dy0v^2)^0.5)));
                                else
                                    theta3v=-pi()-asin(Dy0v/((Dx3v^2+Dy0v^2)^0.5));
                                end
                            end
                            V=0;
                            if (Dx3u^2+Dy0u^2)^0.5<R % Lettieri et al equation 5
                                V=k_v3*((Dx3u^2+Dy0u^2)^0.5)/R;
                            else
                                V=k_v3*((R/((Dx3u^2+Dy0u^2)^0.5))^2);
                            end
                                u(i,j,k)=uNB(i,j,k)+V*sin(theta3u);
                            % Now v
                            if (Dx3v^2+Dy0v^2)^0.5<R % Lettieri et al equation 5
                                V=k_v3*((Dx3v^2+Dy0v^2)^0.5)/R;
                            else
                                V=k_v3*((R/((Dx3v^2+Dy0v^2)^0.5))^2);
                            end
                            v(i,j,k)=vNB(i,j,k)-V*cos(theta3v);
                        end
                        if isreal(u)==0 || isreal(v)==0 || isreal(w)==0 || isreal(p)==0
                            'complex'
                            pause(5)
                            notcomplex=2;
                        else
                            notcomplex=1;
                        end
                    end
                end
            end
        end
    end
   
    % Calculate pressure field
    % First calculate new p_CA, noting that p_0 is proportional to mass of
    % air in the cold aisle
    maxV=(max(max(max(abs(u)))));
    if (max(max(max(abs(v)))))>maxV
        maxV=(max(max(max(abs(v)))));
    end
    if (max(max(max(abs(w)))))>maxV
        maxV=(max(max(max(abs(w)))));
    end
    
    if EquilCond==2 % If p_CA held constant
        if iteration<101
        elseif abs(1-(abs(sum(ContFlowStore(iteration-100:iteration-51))-sum(ContFlowStore(iteration:iteration-50)))/50)/ContFlowStore(iteration))<0.001 % If change in mean bypass from last 50 iterations to the previous 50 is less than 0.1%, i.e. if bypass flow rate is stable
            Weight=0;
            VIn=((1-Weight)*(VServ+ContFlowStore(iteration)+SlotFlowStore(iteration))+Weight*VInStore(iteration-1)); % Calculate supply flow rate required for steady state
        end
        VInStore(iteration)=VIn;
    else % VIn held constant
        if p_CA>0
            p_CA=p_CA*(H*L*Contz+(VIn-ContFlowStore(iteration)-SlotFlowStore(iteration)-VServ)*(0.125*H*L*Contz/VIn))/(H*L*Contz);
        else
            p_CA=p_CA*(H*L*Contz-(VIn-ContFlowStore(iteration)-SlotFlowStore(iteration)-VServ)*(0.125*H*L*Contz/VIn))/(H*L*Contz);
        end
        p_CA_store(iteration)=p_CA;        
    end
    
    for i=1:nx+2
        for j=1:ny+2
            ColdAisle=1; % Baseline pressure differs between hot and cold aisle
            for k=1:nz+2
                if k>kstart
                    ColdAisle=0;
                end
                if ColdAisle==1
                    p(i,j,k,2)=p_0+p_CA-((0.5*(u(i-(i>1),j,k)+u(i-(i==nx+2),j,k)))^2+(0.5*(v(i,j-(j>1),k)+v(i,j-(j==ny+2),k)))^2+(0.5*(w(i,j,k-(k>1))+w(i,j,k-(k==nz+2))))^2); % Apply Bernoulli's equation
                else
                    p(i,j,k,2)=p_0-((0.5*(u(i-(i>1),j,k)+u(i-(i==nx+2),j,k)))^2+(0.5*(v(i,j-(j>1),k)+v(i,j-(j==ny+2),k)))^2+(0.5*(w(i,j,k-(k>1))+w(i,j,k-(k==nz+2))))^2); % Apply Bernoulli's equation
                end
            end
        end
    end
    
    % Calculate Phi residual
    if PhiRes>PhiResTarget
        PhiRes=0;
        PhiSum=0;
        PhiChangeMax=0;
        for i=1:nx+2
            for j=1:ny+2
                for k=1:nz+2
                    if isnan(Phi(i,j,k,PhiIt))==0
                    PhiRes=PhiRes+abs(Phi(i,j,k,PhiIt)-Phi(i,j,k,PhiIt-1));
                    PhiSum=PhiSum+abs(Phi(i,j,k,PhiIt));
                    PhiChange=abs(Phi(i,j,k,PhiIt)-Phi(i,j,k,PhiIt-1));
                        if PhiChange>PhiChangeMax
                            PhiChangeMax=[i j k PhiChange];
                        end
                    end
                end
            end
        end
        if PhiSum==0
            PhiRes=1;
        else
            PhiRes=PhiRes/PhiSum;
            PhiResStore(iteration)=PhiRes;
        end
        if PhiRes<=PhiResTarget
            PhiConvergedIt=iteration;
            U_cube=(6.761-((39.8-20.7)*0.134*1.2*1.005))/(39.8-20.7); % 24/1/17 test, 6.761kW heat load, average T in HA 39.8C, average T_supply=20.7C, V_supply=134l/s
            %U_cube=0;
            T(1:nx+2,1:ny+2,1:kstart,iteration)=(T_0+(EnvHeatLoad*(L*H+H*2*Contz+Contz*L)/(c_p*rho_0*VIn)));
            T_HA=T_0+(EServ+(EnvHeatLoad*(L*H+H*2*Contz+Contz*L)))/(U_cube+VIn*rho_0*c_p);
            T(1:nx+2,1:ny+2,kstart+1:nz+2,iteration)=T_HA;
            U_cube=U_cube/((nx+2)*(ny+2)*(nz+2-kstart)); % Convert U_cube to heat loss per cell in the hot aisle
            U_cubeCA=U_cube*((nx+2)*(ny+2)*kstart)/((nx+2)*(ny+2)*(nz+2-kstart)); % Heat loss per cell in cold aisle
        elseif iteration>200
            if mean(PhiResStore(iteration-50:iteration))>mean(PhiResStore(iteration-100:iteration-51)) % if Phi isn't converging
                if WSOR>1
                    WSOR=WSOR-0.1;
                    if WSOR<1
                        WSOR=1;
                    end
                else
                    'Phi failed to converge'
                    PhiRes
                    PhiConvergedIt=iteration;
                    %U_cube=(6.761-((39.8-20.7)*0.134*1.2*1.005))/(39.8-20.7); % 24/1/17 test, 6.761kW heat load, average T in HA 39.8C, average T_supply=20.7C, V_supply=134l/s
                    U_cube=0;
                    T(1:nx+2,1:ny+2,1:kstart,iteration)=(T_0+(EnvHeatLoad*(L*H+H*2*Contz+Contz*L)/(c_p*rho_0*VIn)));
                    T_HA=T_0+(EServ+(EnvHeatLoad*(L*H+H*2*Contz+Contz*L)))/(U_cube+VIn*rho_0*c_p);
                    T(1:nx+2,1:ny+2,kstart+1:nz+2,iteration)=T_HA;
                    U_cube=U_cube/((nx+2)*(ny+2)*(nz+2-kstart)); % Convert U_cube to heat loss per cell in the hot aisle
                    U_cubeCA=U_cube*((nx+2)*(ny+2)*kstart)/((nx+2)*(ny+2)*(nz+2-kstart)); % Heat loss per cell in cold aisle
                end
            end
        end
    end
    
    % Calculate p residual
        pRes=0;
        pSum=0;
        pChangeMax=0;
        for i=1:nx+2
            for j=1:ny+2
                for k=1:nz+2
                    if isnan(p(i,j,k,2))==0
                        pRes=pRes+abs(p(i,j,k,2)-p(i,j,k,1));
                        pSum=pSum+abs(p(i,j,k,2))-p_0;
                        pChange=abs(p(i,j,k,2)-p(i,j,k,1));
                        if pChange>pChangeMax
                            pChangeMax=[i j k pChange];
                        end
                    end
                end
            end
        end
        if pSum==0 || pRes==0
            pRes=1;
        else
            pRes=pRes/pSum;
            pResStore(iteration)=pRes;
        end
        
        if pRes<=pResTarget && pConvergedIt==0
            pConvergedIt=iteration;
        end
        p(1:nx+2,1:ny+2,1:nz+2,1)=p(1:nx+2,1:ny+2,1:nz+2,2);

iteration=iteration+1;
Phi(1:nx+2,1:ny+2,1:nz+2,1)=Phi(1:nx+2,1:ny+2,1:nz+2,2);
end

TimeEnd=[datestr(now, 'HH') datestr(now, 'MM')]

p=p-101000; % Convert pressures from absolute to gauge

prompt='Show plots? 0 for no, 1 for yes.';
i=input(prompt);
while i==1
    prompt='1 for surface plots, 2 for arrow plots';
    j=input(prompt);
    if j==1
        uvwpplots_thesis(iteration, istore, jstore, kstore, nx, ny, nz, u, v, w, p, T); % call function for plots
    elseif j==2
        ArrowPlot_thesis(iteration, istore, jstore, kstore, nx, ny, nz, u, v, w); % call function for arrow plots
    end
    prompt='More plots? 0 for no, 1 for yes';
    i=input(prompt);
end
for i=1:nx+1
    for j=1:ny+1
        for k=1:nz+1
            V(i,j,k)=(u(i,j,k)^2+v(i,j,k)^2+w(i,j,k)^2)^0.5; % Store velocity magnitude
        end
    end
end

% Calculate flow across outlet
iteration=iteration-1;
Phi1=0;
Phi2=0;
VOut=0;
for j=1:ny+2
    for k=1:nz+2
        if BCs(nx+2,j,k)==3 % outlet
            VOut=VOut+Dy*Dz*(Phi(nx+1,j,k,PhiIt)-Phi(nx+2,j,k,PhiIt))/Dx;
        end
    end
end

% Results for comparison with experimental measurements
ServerInletTemps=0;
startSIT(3)=0;
for i=1:nx+2
    for j=1:ny+2
        for k=1:nz+2
            if BCs(i,j,k)==5
                if startSIT(1)==0
                    startSIT(1)=i;
                    startSIT(2)=j;
                    startSIT(3)=k;
                end
                ServerInletTemps(i-startSIT(1)+1,j-startSIT(2)+1)=T(i,j,k,iteration-1);
            end
        end
    end
end
csvwrite('ServerInletTemps.csv',ServerInletTemps);

% Grid interpolation
[X1,X2,X3]=ndgrid(-Dx/2:Dx:Dx*nx+Dx/2, -Dy/2:Dy:Dy*ny+Dy/2, -Dz/2:Dz:Dz*nz+Dz/2); % T nodes are misaligned with walls
TLast=T(:,:,:,iteration-1);
F=griddedInterpolant(X1,X2,X3,TLast);
% Cold aisle
[X1q,X2q,X3q]=ndgrid(L-2.8:0.8:L-0.4, 0.4:1.33:1.73, 0.666); % Positions for experiments with HA door open
CAResults=F(X1q,X2q,X3q);
% [X1q,X2q,X3q]=ndgrid(L-2.8:0.8:L-0.4, 1.73, 0.7); % Positions for experiments with HA door closed (Jan 2017)
% CAResults=F(X1q,X2q,X3q);
csvwrite('CAResults.csv',CAResults);
% Hot aisle
[X1q,X2q,X3q]=ndgrid(L-2.8:0.8:L-0.4, 0.4:0.65:1.7, D-0.55:-0.4:D-0.95); % Positions for experiments with HA door closed (Jan 2017)
%[X1q,X2q,X3q]=ndgrid(L-2.8:0.8:L-0.4, 0.4:1.33:1.73, D-0.67); % Positions for experiments with HA door open (summer 2016)
HAResults=F(X1q,X2q,X3q);
csvwrite('HAResults.csv',HAResults);
% Racks
[X1q,X2q,X3q]=ndgrid(L-2.8, 1.2:0.4:1.6, Contz+0.53);
Rack1Results=F(X1q,X2q,X3q);
[X1q,X2q,X3q]=ndgrid(L-0.4, 1.2:0.4:1.6, Contz+0.53);
Rack4Results=F(X1q,X2q,X3q);
csvwrite('Rack4Results.csv',Rack4Results);
% LB sides
[X1q,X2q,X3q]=ndgrid(L-2.8-WServ/2-0.0025:WServ+0.005:L-2.8+WServ/2+0.0025, Servy+HServ/8:HServ/4:Servy+HServ, Contz+0.04);
LBSidesRack1Results=F(X1q,X2q,X3q);
[X1q,X2q,X3q]=ndgrid(L-0.4-WServ/2-0.0025:WServ+0.005:L-0.4+WServ/2+0.0025, Servy+HServ/8:HServ/4:Servy+HServ, Contz+0.04);
LBSidesRack4Results=F(X1q,X2q,X3q);
csvwrite('LBSidesRack4Results.csv',LBSidesRack4Results);

%continuity map
for i=2:nx+1
    for j=2:ny+1
        for k=2:nz+1
                Cont(i-1,j-1,k-1)=Phi(i,j,k,PhiIt-1)-(Phi(i+1,j,k,PhiIt-1)+Phi(i-1,j,k,PhiIt-1)+Phi(i,j+1,k,PhiIt-1)+Phi(i,j-1,k,PhiIt-1)+Phi(i,j,k+1,PhiIt-1)+Phi(i,j,k-1,PhiIt-1))/6;
                ContProp(i-1,j-1,k-1)=(Phi(i,j,k,PhiIt-1)-(Phi(i+1,j,k,PhiIt-1)+Phi(i-1,j,k,PhiIt-1)+Phi(i,j+1,k,PhiIt-1)+Phi(i,j-1,k,PhiIt-1)+Phi(i,j,k+1,PhiIt-1)+Phi(i,j,k-1,PhiIt-1))/6)/(Phi(i,j,k-1,PhiIt-1)-Phi(i,j,k+1,PhiIt-1));
        end
    end
end

% calculate flow rate through containment (test)
for k=2:nz+1
    if ((k-1)*Dz-(Dz/2))>Contz && ((k-1)*Dz-(Dz/2))<(Contz+Dz)
        break
    end
end
VCont=0; % Flow through containment
for i=2:nx+1
    for j=2:ny+1
        VCont=VCont+Dx*Dy*(Phi(i,j,k-1,PhiIt-1)-Phi(i,j,k,PhiIt-1))/Dz;
    end
end

% Calculate bypass and recirculation
RecircMap=0;
BypassMap=0;
for i=1:nx+2
    for j=1:ny+2
        if w(i,j,kcont)<0 && BCs(i,j,kcont)~=5
            RecircMap(i,j)=w(i,j,kcont);
        elseif BCs(i,j,kcont)~=5
            BypassMap(i,j)=w(i,j,kcont);
        end
    end
end
TotalBypass=sum(sum(BypassMap))*Dx*Dy;
TotalRecirc=-sum(sum(RecircMap))*Dx*Dy;
%RecircBPMap=RecircMap+BypassMap;

[Results]=SystemPF2(VIn, VServ, EServ, ContFlow, ServerInletTemps, T_0);